/*
    copyright (c) 2025 Tino Mettler

    darktable is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    darktable is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software.  If not, see <http://www.gnu.org/licenses/>.
*/

body {
    width: 100vw;
    height: 100vh;
    display: flex;
    flex-direction: column;
}

.heading h1 {
    text-align: center;
    display: grid;
}

.gallery {
    display: flex;
    flex-wrap: wrap;
    justify-content: center;
    align-items: center;
}

.gallery img {
    cursor: pointer;
    object-fit: contain;
}

.thumb {
    object-fit: contain;
    flex-shrink: 0;
    min-width: 100%;
    min-height: 100%;
    max-width: 100%;
    max-height: 100%;
}

.thumb-box {
    display: flex;
    align-items: center;
    object-fit: contain;
    justify-content: center;
}
